function love.load()
    -- Define how many parts the puzzle should have.
    -- number of parts = NUMBER_OF_ROWS * NUMBER_OF_COLUMNS
    NUMBER_OF_ROWS    = 5
    NUMBER_OF_COLUMNS = 5
    
    -- Define the snapping distance of the grid in pixel.
    SNAPPING_DISTANCE = 20
    
    -- List all the global variables
    video = nil
    video_canvas = nil
    video_playable = nil
    pieces = {}
    active_piece = nil
    piece_width  = nil
    piece_height = nil
end

function love.filedropped(file)
    -- Try using the file as a video.
    --video_playable, video = pcall(love.graphics.newVideo(file))
    video_playable, video = true, love.graphics.newVideo(file)
    
    if video_playable then
        -- Start playing the video.
        video:play()
        
        -- Create video canvas to draw the video frames to.
        video_canvas = love.graphics.newCanvas(video:getWidth(), video:getHeight())
        
        -- Create the puzzle pieces.
        piece_width  = math.floor(video:getWidth() /NUMBER_OF_COLUMNS)
        piece_height = math.floor(video:getHeight()/NUMBER_OF_ROWS)
        create_puzzle_pieces()
    else
        -- Delete the error message from pcall.
        video = nil
    end
end

function love.update(dt)
    -- Play the video on an infinite loop.
    if video ~= nil and not video:isPlaying() then
        video:rewind()
        video:play()
    end
end

function love.keypressed(key)
    if key=="escape" then
        love.event.quit()
    end
end

function love.draw()
    if video == nil then
        -- Show the instructions to start the puzzle.
        if video_playable == nil then
            love.graphics.print("Drag and drop an ogg theora video in here.", 241, 293)
        else
            -- The user tried a file, but it did not work. Change the instructions accordingly.
            love.graphics.print("Drag and drop an ogg theora video in here.\n  This file does not work. Try a different file.", 241, 293)
        end
    else
        -- Draw the grid
        local x_offset = 0.5*(love.graphics.getWidth()  - video:getWidth() )
        local y_offset = 0.5*(love.graphics.getHeight() - video:getHeight())
        
        for row = 1, NUMBER_OF_ROWS do
            for column = 1, NUMBER_OF_COLUMNS do
                love.graphics.line(x_offset + piece_width*(row-1), y_offset + piece_height*(column-1), x_offset + piece_width*row, y_offset + piece_height*(column-1))
                love.graphics.line(x_offset + piece_width*(row-1), y_offset + piece_height*(column-1), x_offset + piece_width*(row-1), y_offset + piece_height*column)
            end
        end
        
        love.graphics.line(x_offset + piece_width*NUMBER_OF_ROWS, y_offset                                 , x_offset + piece_width*NUMBER_OF_ROWS, y_offset + piece_height*NUMBER_OF_COLUMNS)
        love.graphics.line(x_offset                             , y_offset + piece_height*NUMBER_OF_COLUMNS, x_offset + piece_width*NUMBER_OF_ROWS, y_offset + piece_height*NUMBER_OF_COLUMNS)
        
        -- Draw the current video frame
        love.graphics.setCanvas(video_canvas)
            love.graphics.draw(video, 0, 0)
        love.graphics.setCanvas()
        
        -- Sort the pieces, so that the last clicked piece is drawn on top.
        table.sort(pieces, function(a, b) return a.last_active < b.last_active end)
        
        -- Draw all pieces.
        for index, piece in ipairs(pieces) do
            love.graphics.draw(video_canvas, piece.quad, piece.position.x, piece.position.y)
        end
    end
end

function love.mousepressed(x, y, button)
    if video ~= nil and button == 1 then
        -- Look at all pieces under the cursor and find the upper piece
        for index, piece in ipairs(pieces) do
            if x >= piece.position.x and x <= piece.position.x + piece_width and y >= piece.position.y and y <= piece.position.y + piece_height then
                if active_piece == nil then
                    -- If no active_piece is found until now, take the first one under the cursor.
                    active_piece = piece
                elseif active_piece.last_active < piece.last_active then
                    -- If there is an active_piece then use the one nearer to the screen.
                    active_piece = piece                    
                end
            end
        end
        
        -- Update the timer on the active piece, so that it is drawn nearest to the screen.
        if active_piece ~= nil then
            active_piece.last_active = love.timer.getTime()
        end
    end
end

function love.mousemoved(x, y, dx, dy)
    if video ~= nil and active_piece ~= nil then
        -- Move the piece with the mouse.
        active_piece.position.x = active_piece.position.x + dx
        active_piece.position.y = active_piece.position.y + dy
    end
end

function love.mousereleased(x, y)
    local x_offset = 0.5*(love.graphics.getWidth()  - video:getWidth() )
    local y_offset = 0.5*(love.graphics.getHeight() - video:getHeight())
    
    -- If the active piece is near a valid grid position, snap it in place.
    if active_piece ~= nil then
        for row = 1, NUMBER_OF_ROWS do
            for column = 1, NUMBER_OF_COLUMNS do
                if (x_offset + piece_width*(row-1) - active_piece.position.x)^2 + (y_offset + piece_height*(column-1) - active_piece.position.y)^2 <= SNAPPING_DISTANCE^2 then
                    active_piece.position.x = x_offset + piece_width*(row-1)
                    active_piece.position.y = y_offset + piece_height*(column-1)
                end
            end
        end
        
        -- Drop the active piece.
        active_piece = nil
    end
end

function create_puzzle_pieces()
    for row = 1, NUMBER_OF_ROWS do
        for column = 1, NUMBER_OF_COLUMNS do
            local piece = {}
            
            piece.position = {}
            -- Place the new piece somewhere random, but fully, inside the screen
            piece.position.x = love.math.random(0, love.graphics.getWidth()  - piece_width)
            piece.position.y = love.math.random(0, love.graphics.getHeight() - piece_height)
            
            -- This is used to sort the pieces for drawing. The bigger the timestamp, the nearer to the screen the piece is drawn.
            piece.last_active = love.timer.getTime()
            
            -- Define the part of the video from which the piece is taken.
            piece.quad = love.graphics.newQuad((row-1)*piece_width, (column-1)*piece_height, piece_width, piece_height, video:getDimensions())
            
            table.insert(pieces, piece)
        end
    end
end
